import argparse
import copy
import logging
import os
import pprint
import random
import glob
import json
import math
import re

import numpy as np
import torch
from torch.utils.data import (DataLoader, RandomSampler, SequentialSampler)
from torch.utils.data.distributed import DistributedSampler
from tqdm import tqdm, trange

from tensorboardX import SummaryWriter

from transformers import (WEIGHTS_NAME, BertConfig, BertTokenizer)
from transformers import AdamW
try:
    from transformers import WarmupLinearSchedule as get_linear_schedule_with_warmup
except:
    from transformers import get_linear_schedule_with_warmup

from modeling_bert_dst import (BertForDST)
from utils import TensorListDataset, ALL_SLOTS, SPEC_TOKENS, REQUEST_SLOTS, UtterOp, CLASS_TYPES
from preview_convert_data import InputFeatures

# logging.basicConfig(filename='preview.log', level=logging.INFO, filemode='w')
logger = logging.getLogger(__name__)

ALL_MODELS = tuple(BertConfig.pretrained_config_archive_map.keys())

MODEL_CLASSES = {
    'bert': (BertConfig, BertForDST, BertTokenizer),
}


def set_seed(args):
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    if args.n_gpu > 0:
        torch.cuda.manual_seed_all(args.seed)


def to_list(tensor):
    return tensor.detach().cpu().tolist()


def batch_to_device(batch, device):
    batch_on_device = []
    for element in batch:
        if isinstance(element, dict):
            batch_on_device.append({k: v.to(device) for k, v in element.items()})
        else:
            batch_on_device.append(element.to(device))
    return tuple(batch_on_device)


def mask_tokens(inputs, tokenizer, mlm_probability=0.15):
    """ Prepare masked tokens inputs/labels for masked language modeling: 80% MASK, 10% random, 10% original. """
    labels = inputs.clone()
    # We sample a few tokens in each sequence for masked-LM training (with probability args.mlm_probability defaults to 0.15 in Bert/RoBERTa)
    probability_matrix = torch.full(labels.shape, mlm_probability)
    # print(probability_matrix)
    # special_tokens_mask = [tokenizer.get_special_tokens_mask(val, already_has_special_tokens=True) for val in labels.tolist()]
    probability_matrix.masked_fill_(torch.tensor(labels == 0, dtype=torch.bool), value=0.0)
    # print(probability_matrix)
    masked_indices = torch.bernoulli(probability_matrix).bool()
    # print(masked_indices)
    labels[~masked_indices] = -100  # We only compute loss on masked tokens
    # print(labels)
    
    # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
    indices_replaced = torch.bernoulli(torch.full(labels.shape, 0.8)).bool() & masked_indices
    inputs[indices_replaced] = tokenizer._convert_token_to_id("[MASK]")
    
    # 10% of the time, we replace masked input tokens with random word
    indices_random = torch.bernoulli(torch.full(labels.shape, 0.5)).bool() & masked_indices & ~indices_replaced
    random_words = torch.randint(len(tokenizer), labels.shape, dtype=torch.long)
    
    if torch.cuda.is_available():
        inputs[indices_random] = random_words[indices_random].cuda()
    else:
        inputs[indices_random] = random_words[indices_random]
    
    # The rest of the time (10% of the time) we keep the masked input tokens unchanged
    
    return inputs, labels


def train(args, model, tokenizer, mlm_dataset, dataset, schema):
    """ Train the model """
    if args.local_rank in [-1, 0]:
        tb_writer = SummaryWriter()
    
    mlm_train_sampler = RandomSampler(mlm_dataset) if args.local_rank == -1 else DistributedSampler(mlm_dataset)
    mlm_train_dataloader = DataLoader(mlm_dataset, sampler=mlm_train_sampler, batch_size=args.mlm_train_batch_size)
    t_total = len(mlm_train_dataloader) // args.gradient_accumulation_steps * args.num_train_epochs
    if args.save_epochs > 0: args.save_steps = t_total // args.num_train_epochs * args.save_epochs
    num_warmup_steps = int(t_total * args.warmup_proportion)
    
    # Prepare optimizer and schedule (linear warmup and decay)
    no_decay = ['bias', 'LayerNorm.weight']
    optimizer_grouped_parameters = [
        {'params': [p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)],
         'weight_decay': args.weight_decay},
        {'params': [p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
    ]
    optimizer = AdamW(optimizer_grouped_parameters, lr=args.learning_rate, eps=args.adam_epsilon)
    scheduler = get_linear_schedule_with_warmup(optimizer, num_warmup_steps, t_total)
    if args.fp16:
        try:
            from apex import amp
        except ImportError:
            raise ImportError("Please install apex from https://www.github.com/nvidia/apex to use fp16 training.")
        model, optimizer = amp.initialize(model, optimizer, opt_level=args.fp16_opt_level)
    
    # multi-gpu training (should be after apex fp16 initialization)
    model_single_gpu = model
    if args.n_gpu > 1:
        model = torch.nn.DataParallel(model_single_gpu)
    
    # Distributed training (should be after apex fp16 initialization)
    if args.local_rank != -1:
        model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.local_rank],
                                                          output_device=args.local_rank,
                                                          find_unused_parameters=True)
    
    # Train!
    
    logger.info("***** Running training *****")
    logger.info("  Num MLM examples = %d", len(mlm_dataset))
    logger.info("  Num preview examples = %d", len(dataset))
    logger.info("  Num Epochs = %d", args.num_train_epochs)
    logger.info("  MLM batch size per GPU = %d", args.mlm_train_batch_size)
    logger.info("  preview batch size per GPU = %d", args.preview_train_batch_size)
    
    logger.info("  Gradient Accumulation steps = %d", args.gradient_accumulation_steps)
    logger.info("  Total optimization steps = %d", t_total)
    logger.info("  Warmup steps = %d", num_warmup_steps)
    print('start training ...')
    # if continue_from_global_step > 0:
    #     logger.info("Fast forwarding to global step %d to resume training from latest checkpoint...", continue_from_global_step)
    
    global_step = 0
    tr_loss_dict, logging_loss_dict = {'total': 0.0, 'class': 0.0, 'span': 0.0, 'mlm': 0.0}, {'total': 0.0,
                                                                                              'class': 0.0, 'span': 0.0,
                                                                                              'mlm': 0.0}
    model.zero_grad()
    train_iterator = trange(int(args.num_train_epochs), desc="Epoch", disable=args.local_rank not in [-1, 0])
    set_seed(args)  # Added here for reproductibility (even between python 2 and 3)
    
    preview_start = 0
    
    for _ in train_iterator:
        epoch_iterator = tqdm(mlm_train_dataloader, desc="Iteration", disable=args.local_rank not in [-1, 0])
        
        for step, mlm_batch in enumerate(epoch_iterator):
            
            # # If training is continued from a checkpoint, fast forward
            # # to the state of that checkpoint.
            # if global_step < continue_from_global_step:
            #     if (step + 1) % args.gradient_accumulation_steps == 0:
            #         scheduler.step()  # Update learning rate schedule
            #         global_step += 1
            #     continue
            model.train()
            
            if preview_start >= len(dataset):
                preview_start = 0
            preview_batch = dataset[preview_start: preview_start + args.preview_train_batch_size]
            
            preview_batch = batch_to_device(preview_batch, args.device)
            mlm_batch = batch_to_device(mlm_batch, args.device)
            schema = batch_to_device(schema, args.device)
            
            if torch.cuda.is_available():
                mlm_inputs, mlm_labels = mask_tokens(mlm_batch[0].cuda(), tokenizer)
                mlm_inputs, mlm_labels = mlm_inputs.cuda(), mlm_labels.cuda()
            else:
                mlm_inputs, mlm_labels = mask_tokens(mlm_batch[0], tokenizer)
            
            inputs = {
                'mlm_input_ids': mlm_inputs,
                'mlm_input_mask': mlm_batch[1],
                'mlm_segment_ids': mlm_batch[2],
                'masked_lm_labels': mlm_labels,
                'schema_input_ids': schema[0],
                'schema_input_mask': schema[1],
                'schema_segment_ids': schema[2],
                'input_ids': preview_batch[0],
                'input_mask': preview_batch[1],
                'segment_ids': preview_batch[2],
                'span_label_id': preview_batch[3],
                'class_label_id': preview_batch[4],
            }
            loss, total_class_loss, total_span_loss, masked_lm_loss = model.preview(**inputs)
            
            if args.n_gpu > 1:
                loss = loss.mean()  # mean() to average on multi-gpu parallel (not distributed) training
            if args.gradient_accumulation_steps > 1:
                loss = loss / args.gradient_accumulation_steps
            
            if args.fp16:
                with amp.scale_loss(loss, optimizer) as scaled_loss:
                    scaled_loss.backward()
                torch.nn.utils.clip_grad_norm_(amp.master_params(optimizer), args.max_grad_norm)
            else:
                loss.backward()
                torch.nn.utils.clip_grad_norm_(model.parameters(), args.max_grad_norm)
            
            tr_loss_dict['total'] += loss.item()
            tr_loss_dict['span'] += total_span_loss.item()
            tr_loss_dict['class'] += total_class_loss.item()
            tr_loss_dict['mlm'] += masked_lm_loss.item()
            
            if (step + 1) % args.gradient_accumulation_steps == 0:
                optimizer.step()
                scheduler.step()  # Update learning rate schedule
                model.zero_grad()
                global_step += 1
                
                # Log metrics
                if args.local_rank in [-1, 0] and args.logging_steps > 0 and global_step % args.logging_steps == 0:
                    tb_writer.add_scalar('lr', scheduler.get_last_lr()[0], global_step)
                    tb_writer.add_scalar('total loss',
                                         (tr_loss_dict['total'] - logging_loss_dict['total']) / args.logging_steps,
                                         global_step)
                    tb_writer.add_scalar('class loss',
                                         (tr_loss_dict['class'] - logging_loss_dict['class']) / args.logging_steps,
                                         global_step)
                    tb_writer.add_scalar('span loss',
                                         (tr_loss_dict['span'] - logging_loss_dict['span']) / args.logging_steps,
                                         global_step)
                    tb_writer.add_scalar('mlm loss',
                                         (tr_loss_dict['mlm'] - logging_loss_dict['mlm']) / args.logging_steps,
                                         global_step)
                    logging_loss_dict = copy.deepcopy(tr_loss_dict)
                
                # Save model checkpoint
                if args.local_rank in [-1, 0] and args.save_steps > 0 and global_step % args.save_steps == 0:
                    output_dir = os.path.join(args.output_dir, 'checkpoint-{}'.format(global_step))
                    if not os.path.exists(output_dir):
                        os.makedirs(output_dir)
                    model_to_save = model.module if hasattr(model,
                                                            'module') else model  # Take care of distributed/parallel training
                    model_to_save.save_pretrained(output_dir)
                    tokenizer.save_pretrained(output_dir)
                    torch.save(args, os.path.join(output_dir, 'training_args.bin'))
                    logger.info("Saving model checkpoint to %s", output_dir)
            
            if args.max_steps > 0 and global_step > args.max_steps:
                epoch_iterator.close()
                break
        
        # if args.local_rank == -1 and args.evaluate_during_training:  # Only evaluate when single GPU otherwise metrics may not average well
        #     results = evaluate(args, model_single_gpu, tokenizer, prefix=global_step)
        #     for key, value in results.items():
        #         tb_writer.add_scalar('eval_{}'.format(key), value, global_step)
        
        if args.max_steps > 0 and global_step > args.max_steps:
            train_iterator.close()
            break
    
    if args.local_rank in [-1, 0]:
        tb_writer.close()
    
    return global_step, {k: v / global_step for k, v in tr_loss_dict.items()}


def load_and_cache_examples(args, model, tokenizer):
    if args.local_rank not in [-1, 0]:
        torch.distributed.barrier()  # Make sure only the first process in distributed training process the dataset, and the others will use the cache
    
    # Load data features from cache or dataset file
    
    #     mlm_data_feature = torch.load('mlm_data_feature')
    mlm_data_feature = torch.load('preview/mlm_data_feature_sysdelex')
    preview_data_feature = torch.load('preview/preview_data_feature')
    schema_feat_dict = torch.load('preview/schema_feat_dict')
    
    if args.local_rank == 0:
        torch.distributed.barrier()  # Make sure only the first process in distributed training process the dataset, and the others will use the cache
    
    # Convert to Tensors and build dataset
    mlm_input_ids = torch.tensor([f.input_ids for f in mlm_data_feature], dtype=torch.long)
    mlm_input_mask = torch.tensor([f.input_mask for f in mlm_data_feature], dtype=torch.long)
    mlm_segment_ids = torch.tensor([f.segment_ids for f in mlm_data_feature], dtype=torch.long)
    mlm_example_index = torch.arange(mlm_input_ids.size(0), dtype=torch.long)
    mlm_dataset = TensorListDataset(mlm_input_ids, mlm_input_mask, mlm_segment_ids, mlm_example_index)
    
    schema_input_ids = torch.tensor(
        [schema_feat_dict[slot]['input_ids'] for slot in ALL_SLOTS],
        dtype=torch.long)
    schema_input_mask = torch.tensor(
        [schema_feat_dict[slot]['input_mask'] for slot in ALL_SLOTS],
        dtype=torch.long)
    schema_segment_ids = torch.tensor(
        [schema_feat_dict[slot]['segment_ids'] for slot in ALL_SLOTS],
        dtype=torch.long)
    schema = (schema_input_ids, schema_input_mask, schema_segment_ids)
    
    all_input_ids = torch.tensor([f.input_ids for f in preview_data_feature], dtype=torch.long)
    all_input_mask = torch.tensor([f.input_mask for f in preview_data_feature], dtype=torch.long)
    all_segment_ids = torch.tensor([f.segment_ids for f in preview_data_feature], dtype=torch.long)
    all_example_index = torch.arange(all_input_ids.size(0), dtype=torch.long)
    f_span_label_ids = [f.span_label_id for f in preview_data_feature]
    f_class_label_ids = [f.class_label_id for f in preview_data_feature]
    
    all_span_label_ids = {}
    all_class_label_ids = {}
    for s in model.slot_list:
        all_span_label_ids[s] = torch.tensor([f[s] for f in f_span_label_ids], dtype=torch.float32)
        all_class_label_ids[s] = torch.tensor([f[s] for f in f_class_label_ids], dtype=torch.long)
    dataset = TensorListDataset(all_input_ids, all_input_mask, all_segment_ids,
                                all_span_label_ids,
                                all_class_label_ids,
                                all_example_index)
    logger.info('mlm dataset %d' % len(mlm_dataset))
    logger.info('preview dataset %d' % len(dataset))
    
    return mlm_dataset, dataset, schema


def main():
    parser = argparse.ArgumentParser()
    
    # Required parameters
    parser.add_argument("--task_name", default=None, type=str, required=False,
                        help="Name of the task (e.g., multiwoz21).")
    parser.add_argument("--data_dir", default=None, type=str, required=False,
                        help="Task database.")
    parser.add_argument("--dataset_config", default=None, type=str, required=False,
                        help="Dataset configuration file.")
    parser.add_argument("--predict_type", default=None, type=str, required=False,
                        help="Portion of the data to perform prediction on (e.g., dev, test).")
    parser.add_argument("--model_type", default="bert", type=str, required=False,
                        help="Model type selected in the list: " + ", ".join(MODEL_CLASSES.keys()))
    parser.add_argument("--model_name_or_path", default=None, type=str, required=False,
                        help="Path to pre-trained model or shortcut name selected in the list: " + ", ".join(
                            ALL_MODELS))
    parser.add_argument("--output_dir", default=None, type=str, required=False,
                        help="The output directory where the model checkpoints and predictions will be written.")
    
    # Other parameters
    parser.add_argument("--config_name", default="", type=str,
                        help="Pretrained config name or path if not the same as model_name")
    parser.add_argument("--tokenizer_name", default="", type=str,
                        help="Pretrained tokenizer name or path if not the same as model_name")
    
    parser.add_argument("--max_seq_length", default=384, type=int,
                        help="Maximum input length after tokenization. Longer sequences will be truncated, shorter ones padded.")
    parser.add_argument("--do_train", action='store_true',
                        help="Whether to run training.")
    parser.add_argument("--do_eval", action='store_true',
                        help="Whether to run eval on the <predict_type> set.")
    parser.add_argument("--evaluate_during_training", action='store_true',
                        help="Rul evaluation during training at each logging step.")
    parser.add_argument("--do_lower_case", action='store_true',
                        help="Set this flag if you are using an uncased model.")
    
    parser.add_argument("--dropout_rate", default=0.2, type=float,
                        help="Dropout rate for BERT representations.")
    parser.add_argument("--heads_dropout", default=0.0, type=float,
                        help="Dropout rate for classification heads.")
    parser.add_argument("--class_loss_ratio", default=0.8, type=float,
                        help="The ratio applied on class loss in total loss calculation. "
                             "Should be a value in [0.0, 1.0]. "
                             "The ratio applied on token loss is (1-class_loss_ratio)/2. "
                             "The ratio applied on refer loss is (1-class_loss_ratio)/2.")
    parser.add_argument("--token_loss_for_nonpointable", action='store_true',
                        help="Whether the token loss for classes other than copy_value contribute towards total loss.")
    parser.add_argument("--refer_loss_for_nonpointable", action='store_true',
                        help="Whether the refer loss for classes other than refer contribute towards total loss.")
    
    parser.add_argument("--append_history", action='store_true',
                        help="Whether or not to append the dialog history to each turn.")
    parser.add_argument("--use_history_labels", action='store_true',
                        help="Whether or not to label the history as well.")
    parser.add_argument("--swap_utterances", action='store_true',
                        help="Whether or not to swap the turn utterances (default: sys|usr, swapped: usr|sys).")
    parser.add_argument("--label_value_repetitions", action='store_true',
                        help="Whether or not to label values that have been mentioned before.")
    parser.add_argument("--delexicalize_sys_utts", action='store_true',
                        help="Whether or not to delexicalize the system utterances.")
    parser.add_argument("--class_aux_feats_inform", action='store_true',
                        help="Whether or not to use the identity of informed slots as auxiliary featurs for class prediction.")
    parser.add_argument("--class_aux_feats_ds", action='store_true',
                        help="Whether or not to use the identity of slots in the current dialog state as auxiliary featurs for class prediction.")
    
    parser.add_argument("--mlm_train_batch_size", default=24, type=int,
                        help="Batch size per GPU/CPU for training.")
    parser.add_argument("--preview_train_batch_size", default=4, type=int,
                        help="Batch size per GPU/CPU for evaluation.")
    
    parser.add_argument("--learning_rate", default=1e-5, type=float,
                        help="The initial learning rate for Adam.")
    parser.add_argument('--gradient_accumulation_steps', type=int, default=1,
                        help="Number of updates steps to accumulate before performing a backward/update pass.")
    parser.add_argument("--weight_decay", default=0.0, type=float,
                        help="Weight deay if we apply some.")
    parser.add_argument("--adam_epsilon", default=1e-8, type=float,
                        help="Epsilon for Adam optimizer.")
    parser.add_argument("--max_grad_norm", default=1.0, type=float,
                        help="Max gradient norm.")
    parser.add_argument("--num_train_epochs", default=3.0, type=float,
                        help="Total number of training epochs to perform.")
    parser.add_argument("--max_steps", default=-1, type=int,
                        help="If > 0: set total number of training steps to perform. Override num_train_epochs.")
    parser.add_argument("--warmup_proportion", default=0.0, type=float,
                        help="Linear warmup over warmup_proportion * steps.")
    parser.add_argument("--svd", default=0.0, type=float,
                        help="Slot value dropout ratio (default: 0.0)")
    
    parser.add_argument('--logging_steps', type=int, default=50,
                        help="Log every X updates steps.")
    parser.add_argument('--save_steps', type=int, default=0,
                        help="Save checkpoint every X updates steps. Overwritten by --save_epochs.")
    parser.add_argument('--save_epochs', type=int, default=1,
                        help="Save checkpoint every X epochs. Overrides --save_steps.")
    parser.add_argument("--eval_all_checkpoints", action='store_true',
                        help="Evaluate all checkpoints starting with the same prefix as model_name ending and ending with step number")
    parser.add_argument("--no_cuda", action='store_true',
                        help="Whether not to use CUDA when available")
    parser.add_argument('--overwrite_output_dir', action='store_true',
                        help="Overwrite the content of the output directory")
    parser.add_argument('--overwrite_cache', action='store_true',
                        help="Overwrite the cached training and evaluation sets")
    parser.add_argument('--seed', type=int, default=42,
                        help="random seed for initialization")
    parser.add_argument('--phase', type=str, default="preview",
                        help="different phase of training")
    parser.add_argument("--local_rank", type=int, default=-1,
                        help="local_rank for distributed training on gpus")
    parser.add_argument('--fp16', action='store_true',
                        help="Whether to use 16-bit (mixed) precision (through NVIDIA apex) instead of 32-bit")
    parser.add_argument('--fp16_opt_level', type=str, default='O1',
                        help="For fp16: Apex AMP optimization level selected in ['O0', 'O1', 'O2', and 'O3']."
                             "See details at https://nvidia.github.io/apex/amp.html")
    
    args = parser.parse_args()
    
    assert (args.warmup_proportion >= 0.0 and args.warmup_proportion <= 1.0)
    assert (args.svd >= 0.0 and args.svd <= 1.0)
    #     assert(args.class_aux_feats_ds is False or args.per_gpu_eval_batch_size == 1)
    #     assert(not args.class_aux_feats_inform or args.per_gpu_eval_batch_size == 1)
    #     assert(not args.class_aux_feats_ds or args.per_gpu_eval_batch_size == 1)
    
    dst_slot_list = ALL_SLOTS
    dst_class_types = CLASS_TYPES
    dst_class_labels = len(dst_class_types)
    
    # Setup CUDA, GPU & distributed training
    if args.local_rank == -1 or args.no_cuda:
        device = torch.device("cuda" if torch.cuda.is_available() and not args.no_cuda else "cpu")
        args.n_gpu = torch.cuda.device_count()
    else:  # Initializes the distributed backend which will take care of sychronizing nodes/GPUs
        torch.cuda.set_device(args.local_rank)
        device = torch.device("cuda", args.local_rank)
        torch.distributed.init_process_group(backend='nccl')
        args.n_gpu = 1
    args.device = device
    
    # Setup logging
    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s',
                        datefmt='%m/%d/%Y %H:%M:%S',
                        level=logging.INFO if args.local_rank in [-1, 0] else logging.WARN)
    logger.warning("Process rank: %s, device: %s, n_gpu: %s, distributed training: %s, 16-bits training: %s",
                   args.local_rank, device, args.n_gpu, bool(args.local_rank != -1), args.fp16)
    
    # Set seed
    set_seed(args)
    
    # Load pretrained model and tokenizer
    if args.local_rank not in [-1, 0]:
        torch.distributed.barrier()  # Make sure only the first process in distributed training will download model & vocab
    
    args.model_type = args.model_type.lower()
    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(args.config_name if args.config_name else args.model_name_or_path)
    
    # Add DST specific parameters to config
    config.dst_dropout_rate = args.dropout_rate
    config.dst_heads_dropout_rate = args.heads_dropout
    config.dst_class_loss_ratio = args.class_loss_ratio
    config.dst_token_loss_for_nonpointable = args.token_loss_for_nonpointable
    config.dst_refer_loss_for_nonpointable = args.refer_loss_for_nonpointable
    config.dst_class_aux_feats_inform = args.class_aux_feats_inform
    config.dst_class_aux_feats_ds = args.class_aux_feats_ds
    config.dst_slot_list = dst_slot_list
    config.dst_class_types = dst_class_types
    config.dst_class_labels = dst_class_labels
    
    if 'phase' in args:
        config.phase = args.phase
    else:
        config.phase = 'preview'
    
    logger.info("Training/evaluation parameters %s", args)
    logger.info("Updated model config: %s" % config)
    
    # Preview Training
    
    # tokenizer = tokenizer_class.from_pretrained(args.tokenizer_name if args.tokenizer_name else args.model_name_or_path, do_lower_case=args.do_lower_case)
    model = model_class.from_pretrained(args.model_name_or_path, from_tf=bool('.ckpt' in args.model_name_or_path),
                                        config=config)
    special_tokens_dict = {'additional_special_tokens': SPEC_TOKENS + list(REQUEST_SLOTS.values())}
    tokenizer = tokenizer_class.from_pretrained(
        args.tokenizer_name if args.tokenizer_name else args.model_name_or_path,
        do_lower_case=True, do_basic_tokenize=True)
    tokenizer.add_special_tokens(special_tokens_dict)  # 记得 TODO model.resize
    
    model.resize_token_embeddings(len(tokenizer))
    model.resize(len(tokenizer))
    
    if args.local_rank == 0:
        torch.distributed.barrier()  # Make sure only the first process in distributed training will download model & vocab
    model.to(args.device)
    
    mlm_dataset, dataset, schema = load_and_cache_examples(args, model, tokenizer)
    
    # # If output files already exists, assume to continue training from latest checkpoint (unless overwrite_output_dir is set)
    # continue_from_global_step = 0 # If set to 0, start training from the beginning
    # if os.path.exists(args.output_dir) and os.listdir(args.output_dir) and args.do_train and not args.overwrite_output_dir:
    #     checkpoints = list(os.path.dirname(c) for c in sorted(glob.glob(args.output_dir + '/*/' + WEIGHTS_NAME, recursive=True)))
    #     if len(checkpoints) > 0:
    #         checkpoint = checkpoints[-1]
    #         logger.info("Resuming training from the latest checkpoint: %s", checkpoint)
    #         continue_from_global_step = int(checkpoint.split('-')[-1])
    #         model = model_class.from_pretrained(checkpoint)
    #         model.to(args.device)
    
    global_step, tr_loss_dic = train(args, model, tokenizer, mlm_dataset, dataset, schema)
    logger.info(" global_step = %s, average loss = %s", global_step, str(tr_loss_dic))


if __name__ == "__main__":
    main()
